// Copyright (c) 2017, Ravi Teja Gudapati. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

library jaguar_reflect.base;

import 'dart:async';
import 'package:jaguar/jaguar.dart' as j;
import 'dart:mirrors';
import 'package:quiver/core.dart';

part 'route.dart';

part 'inputs/inject.dart';
part 'inputs/input.dart';
part 'interceptor/interceptor.dart';

part 'make_param.dart';
part 'path_variable.dart';
part 'query_param.dart';

part 'interceptor/wrapper/reflected_wrapper.dart';
part 'interceptor/wrapper/wrapper_maker.dart';

class JaguarReflected implements j.RequestHandler {
  final dynamic _handler;

  final List<ReflectedRoute> _routes = <ReflectedRoute>[];

  JaguarReflected(this._handler) {
    _parseApi();
  }

  Future<j.Response> handleRequest(j.Request req, {String prefix: ''}) async {
    for (ReflectedRoute route in _routes) {
      j.Response response =
      await route.handleRequest(req, prefix: prefix + route.prefix);
      if (response is j.Response) {
        return response;
      }
    }

    return null;
  }

  void _parseApi() {
    _routes.clear();

    InstanceMirror im = reflect(_handler);
    final List<j.Api> apis = im.type.metadata
        .map((InstanceMirror aim) => aim.reflectee)
        .where((dynamic ref) => ref is j.Api)
        .toList() as List<j.Api>;
    if (apis.length == 0) {
      throw new Exception('Handler is not decorated with Api!');
    }

    final List wrappers = _detectWrappers(im.type.metadata);

    _parse(im, apis.first.url, topWrapper: wrappers);
  }

  void _parse(InstanceMirror im, String pathPrefix,
      {List topWrapper: const []}) {
    im.type.declarations.forEach((Symbol s, DeclarationMirror decl) {
      if (decl.isPrivate) return;

      // Collect groups
      if (decl is VariableMirror) {
        final List<j.Group> groups = decl.metadata
            .where((InstanceMirror annot) => annot.reflectee is j.Group)
            .map((InstanceMirror annot) => annot.reflectee)
            .toList() as List<j.Group>;

        if (groups.length == 0) return;

        if (!decl.isFinal) {
          throw new Exception('Group must be final!');
        }

        InstanceMirror gim = im.getField(s);
        dynamic group = gim.reflectee;
        if (group == null) {
          throw new Exception('Group cannot be null!');
        }

        List<j.RouteGroup> rg = gim.type.metadata
            .where((InstanceMirror annot) => annot.reflectee is j.RouteGroup)
            .map((InstanceMirror annot) => annot.reflectee)
            .toList() as List<j.RouteGroup>;

        if (rg.length == 0) {
          throw new Exception('Group must be annotated with RouteGroup!');
        }

        final List<j.RouteWrapper> wrappers = gim.type.metadata
            .where((InstanceMirror annot) => annot.reflectee is j.RouteWrapper)
            .map((InstanceMirror annot) => annot.reflectee)
            .toList() as List<j.RouteWrapper>;

        _parse(gim, pathPrefix + groups.first.path + rg.first.path,
            topWrapper: wrappers);
      }

      //Collect routes

      if (decl is! MethodMirror) return;

      final List<j.RouteBase> routes = decl.metadata
          .where((InstanceMirror annot) => annot.reflectee is j.RouteBase)
          .map((InstanceMirror annot) => annot.reflectee)
          .toList() as List<j.RouteBase>;
      if (routes.length == 0) return;

      final List wrappers = _detectWrappers(decl.metadata);

      wrappers.insertAll(0, topWrapper);

      InstanceMirror method = im.getField(s);

      routes
          .map((j.RouteBase route) => new ReflectedRoute.buildForClass(
              method.reflectee, route, pathPrefix, wrappers, im))
          .forEach(_routes.add);
    });
  }
}

JaguarReflected reflectJaguar(Object routeGroup) =>
    new JaguarReflected(routeGroup);

List _detectWrappers(List<InstanceMirror> annots) {
  final List wrappers = [];

  for(InstanceMirror annot in annots) {
    dynamic ref = annot.reflectee;
    if(ref is j.RouteWrapper) {
      wrappers.add(ref);
    } else if(ref is j.Wrap) {
      wrappers.addAll(ref.interceptorChain);
    }
  }

  return wrappers;
}
