part of jaguar_reflect.base;

/// A Reflected route. Contains all the information to handle a request for the
/// route
///
/// Call [handleRequest] to handle requests for this route
class ReflectedRoute implements j.RequestHandler {
  /// The route info
  final j.RouteBase route;

  /// Prefix to the route
  final String prefix;

  /// The handler method, function or closure
  final Function handler;

  /// Used to invoke the handler
  final ClosureMirror _invoker;

  /// Injections for required parameters to route handler
  final List<Inject> _required;

  /// Injections for named parameters to route handler
  final Set<RouteQueryParam> _optional;

  /// RouteWrappers to wrap the route with interceptors
  final List<Wrapper> wrappers;

  ReflectedRoute(this.route, this.prefix, this.handler, this._invoker,
      this._required, this._optional, this.wrappers);

  /// Handles request
  Future<j.Response> handleRequest(j.Request request, {String prefix}) async {
    final ctx = new j.ContextImpl(request);

    if (!route.match(
        request.uri.path, request.method, prefix, ctx.pathParams)) {
      return null;
    }

    j.Response response = new j.Response(null,
        statusCode: route.statusCode, headers: route.headers);
    response.headers.mimeType = route.mimeType;
    response.headers.charset = route.charset;

    final List<ReflectedInterceptor> inters = [];
    try {
      // Interceptor Pre
      for (Wrapper wrapper in wrappers) {
        // Create interceptor
        final j.RouteWrapper w = await wrapper.getWrapper(ctx, response);
        final j.Interceptor i = w.createInterceptor();
        final InstanceMirror im = reflect(i);
        final inter = new ReflectedInterceptor(wrapper, i, im);
        inters.add(inter);

        if (inter.wrapper.pre == null) continue;

        // Call Pre
        dynamic result = await inter.invokePre(ctx, response);

        ctx.addOutput(w.interceptorType, w.id, i, result);
      }

      // Call the route
      {
        // Prepare parameters for Route method
        final List reqParam = [];
        for (Inject inj in _required) {
          reqParam.add(_makeParam(inj, ctx, response));
        }

        final Map<Symbol, dynamic> optParams = {};
        for (RouteQueryParam p in _optional) {
          if(ctx.queryParams.containsKey(p.key)) {
            optParams[new Symbol(p.key)] = castQueryParam(p, ctx.queryParams);
          }
        }

        // Call route
        dynamic result = _invoker.apply(reqParam, optParams).reflectee;
        if (result is Future) {
          result = await result;
        }
        if (result is j.Response) {
          response = result;
        } else {
          response.value = result;
        }
      }

      // Interceptor Post
      for (ReflectedInterceptor inter in inters.reversed) {
        if (inter.wrapper.post == null) continue;

        // Call Post
        dynamic resp = await inter.invokePost(ctx, response);
        if (resp is j.Response) {
          response = resp;
        }
      }
    } catch (e) {
      // Call exception handlers of interceptors
      for (ReflectedInterceptor inter in inters.reversed) {
        await inter.inter.onException();
      }
      rethrow;
    }

    return response;
  }

  /// Builds the route for function based routes
  factory ReflectedRoute.build(Function handler, j.RouteBase jRoute,
      String prefix, List<j.RouteWrapper> wrappers) {
    final InstanceMirror im = reflect(handler);

    if (im is! ClosureMirror) {
      throw new Exception('Route handler must be a closure or function!');
    }

    final ClosureMirror c = im as ClosureMirror;
    final MethodMirror m = c.function;

    final List<Inject> required = _parseReqParamsForRoute(m);

    final Set<RouteQueryParam> optional = _parseOptParamsForRoute(m);

    final List<Wrapper> interceptors = wrappers
        .map((j.RouteWrapper inter) => new InstanceWrapper.build(inter))
        .toList();

    return new ReflectedRoute(
        jRoute, prefix, handler, c, required, optional, interceptors);
  }

  /// Builds the route for class based routes
  factory ReflectedRoute.buildForClass(Function handler, j.RouteBase jRoute,
      String prefix, List wraps, InstanceMirror groupIm) {
    final InstanceMirror im = reflect(handler);

    if (im is! ClosureMirror) {
      throw new Exception('Route handler must be a closure or function!');
    }

    final ClosureMirror c = im as ClosureMirror;
    final MethodMirror m = c.function;

    final List<Inject> required = _parseReqParamsForRoute(m);

    final Set<RouteQueryParam> optional = _parseOptParamsForRoute(m);

    final List<Wrapper> interceptors = [];

    for (dynamic wrap in wraps) {
      if (wrap is Symbol) {
        interceptors.add(new MethodWrapper.build(wrap, groupIm));
      } else if (wrap is j.RouteWrapper) {
        interceptors.add(new InstanceWrapper.build(wrap));
      } else {
        throw new Exception("Bug: Invalid wrap source!");
      }
    }

    return new ReflectedRoute(
        jRoute, prefix, handler, c, required, optional, interceptors);
  }
}
